// if you're actually interested in reusing this class for your app
// it'd be a good idea to get in contact with me, Mime Cuvalo: mimecuvalo@gmail.com

function ftpMozilla(observer) {
  this.transportService = Components.classes["@mozilla.org/network/socket-transport-service;1"].getService(Components.interfaces.nsISocketTransportService);
  this.proxyService     = Components.classes["@mozilla.org/network/protocol-proxy-service;1"].getService  (Components.interfaces.nsIProtocolProxyService);
  this.cacheService     = Components.classes["@mozilla.org/network/cache-service;1"].getService           (Components.interfaces.nsICacheService);
  this.observer         = observer;

  var self = this;
  var func = function() { self.keepAlive(); };
  setTimeout(func, 60000);
}

ftpMozilla.prototype = {
  // begin: variables you can set
  host                 : "",
  port                 : 21,
  security             : "",
  login                : "",
  password             : "",
  passiveMode          : true,
  initialPath          : "",             // path we go to first onload

  asciiFiles           : new Array(),    // set to the list of extensions we treat as ASCII files when transfering
  fileMode             : 0,              // 0 == auto, 1 == binary, 2 == ASCII
  hiddenMode           : false,          // show hidden files if true
  ipType               : "IPv4",         // right now, either IPv4 or IPv6
  keepAliveMode        : true,           // keep the connection alive with NOOP's
  networkTimeout       : 30,             // how many seconds b/f we consider the connection to be stale and dead
  proxyHost            : "",
  proxyPort            : 0,
  proxyType            : "",
  reconnectAttempts    : 40,             // how many times we should try reconnecting
  reconnectInterval    : 10,             // number of seconds in b/w reconnect attempts
  reconnectMode        : true,           // true if we want to attempt reconnecting
  sessionsMode         : true,           // true if we're caching directory data

  debug                : function() { }, // set to function that will do something with exception messages
  error                : function() { }, // set to function that will deal with errors
  appendLog            : function() { }, // set to function that will deal with log

  errorConnectStr      : "Unable to make a connection.  Please try again.",  // set to error msg that you'd like to show for a connection error
  passNotShown         : "(password not shown)",                             // set to text you'd like to show in place of password
  // end: variables you can set

  // variables used internally
  isConnected          : false,          // are we connected?
  isReady              : false,          // are we busy writing/reading the control socket?
  isReconnecting       : false,          // are we attempting a reconnect?
  legitClose           : true,           // are we the ones initiating the close or is it a network error
  reconnectsLeft       : 0,              // how many times more to try reconnecting
  networkTimeoutID     : 0,              // a counter increasing with each read and write
  transferID           : 0,              // a counter increasing with each transfer

  controlTransport     : null,
  controlInstream      : null,
  controlOutstream     : null,

  eventQueue           : new Array(),    // commands to be sent
  trashQueue           : new Array(),    // once commands are read, throw them away here b/c we might have to recycle these if there is an error
  doingCmdBatch        : false,

  dataSocket           : null,

  listData             : new Array,      // holds data directory data from the LIST command

  welcomeMessage       : "",             // hello world
  connectedHost        : "",             // name of the host we connect to plus username
  protPSuccess         : false,          // if we were successful in handshaking a secure connection
  localRefreshLater    : '',
  remoteRefreshLater   : '',
  transferMode         : "",             // either "A" or "I"
  currentWorkingDir    : "",             // directory that we're currently, uh, working with
  version              : "0.94.3",  // version of this class - used to avoid collisions in cache
  remoteMonths         : "Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec",         // used in parsing months from list data
  l10nMonths           : new Array("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"), // used in display localized months

  connect : function(reconnect) {
    if (!reconnect) {                                                            // this is not a reconnection attempt
      this.isReconnecting = false;
      this.reconnectsLeft = parseInt(this.reconnectAttempts);

      if (!this.reconnectsLeft || this.reconnectsLeft < 1) {
        this.reconnectsLeft = 1;
      }
    }

    if (!this.eventQueue.length || this.eventQueue[0].cmd != "welcome") {
      this.unshiftEventQueue("welcome", "", "");                                 // wait for welcome message first
    }

    ++this.networkTimeoutID;                                                     // just in case we have timeouts from previous connection
    ++this.transferID;

    try {                                                                        // create a control socket
      var proxyInfo = null;

      if (this.proxyType != "") {                                                // use a proxy
        proxyInfo = this.proxyService.newProxyInfo(this.proxyType, this.proxyHost, this.proxyPort, 0, 30, null);
      }

      if (this.security == "ssl") {                                              // thanks to Scott Bentley. he's a good man, Jeffrey. and thorough.
        this.controlTransport = this.transportService.createTransport(["ssl"],      1, this.host, parseInt(this.port), proxyInfo);
      } else if (!this.security) {
        this.controlTransport = this.transportService.createTransport(null,         0, this.host, parseInt(this.port), proxyInfo);
      } else {
        this.controlTransport = this.transportService.createTransport(["starttls"], 1, this.host, parseInt(this.port), proxyInfo);
      }

      this.controlOutstream = this.controlTransport.openOutputStream(0, 0, 0);
      var controlStream     = this.controlTransport.openInputStream(0, 0, 0);
      this.controlInstream  = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
      this.controlInstream.init(controlStream);

      var self = this;
      var dataListener = {                                                       // async data listener for the control socket
        data            : "",

        onStartRequest  : function(request, context) { },

        onStopRequest   : function(request, context, status) {
          if (!self.isConnected) {                                               // no route to host
            self.appendLog(self.errorConnectStr, 'error', "error");
          }

          self.isConnected = false;

          if (self.dataSocket) {
            self.dataSocket.kill();
          }

          if (self.observer) {
            self.observer.onDisconnected();
            self.observer.onIsReadyChange(true);
          }

          if (!self.legitClose && self.reconnectMode) {                          // try reconnecting
            self.transferMode         = "";
            self.isReconnecting       = true;

            if (self.observer) {
              self.observer.onReconnecting();
            }

            var func = function() { self.reconnect(); };
            setTimeout(func, self.reconnectInterval * 1000);
          } else {
            self.cleanup();
          }
        },

        onDataAvailable : function(request, context, inputStream, offset, count) {
          this.data = self.controlInstream.read(count);                          // read data
          self.readControl(this.data);
        }
      };

      var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"].createInstance(Components.interfaces.nsIInputStreamPump);
      pump.init(controlStream, -1, -1, 0, 0, false);
      pump.asyncRead(dataListener,null);

    } catch(ex) {
      this.debug(ex);
      this.error(this.errorConnectStr);
      this.cleanup();

      if (this.observer) {
        this.observer.onDisconnected();
      }

      return;
    }
  },

  reconnect : function()  {                                                      // ahhhh! our precious connection has been lost,
    if (!this.isReconnecting) {                                                  // must...get it...back...our...precious
      return;
    }

    --this.reconnectsLeft;

    if (this.reconnectsLeft < 0) {
      return;
    }

    this.connect(true);
  },

  disconnect : function() {
    this.legitClose = true;                                                      // this close() is ok, don't try to reconnect
    this.cleanup();

    if (!(!this.isConnected && this.eventQueue.length && this.eventQueue[0].cmd == "welcome")) {
      this.unshiftEventQueue("goodbye", "", "");
      this.writeControl("QUIT");
    }

    if (this.dataSocket) {
      this.dataSocket.kill();
    }

    this.kill();
  },

  cleanup : function() {
    this.eventQueue         = new Array();
    this.trashQueue         = new Array();
    this.transferMode       = "";
    this.currentWorkingDir  = "";
    this.localRefreshLater  = "";
    this.remoteRefreshLater = "";
    ++this.networkTimeoutID;
    ++this.transferID;
  },

  kill : function() {
    try {
      this.controlInstream.close();
    } catch(ex) {
      this.debug(ex);
    }

    try {
      this.controlOutstream.close();
    } catch(ex) {
      this.debug(ex);
    }
  },

  abort : function() {
    this.isReconnecting     = false;

    if (this.dataSocket) {
      this.dataSocket.progressEventSink.bytesTotal = 0;                          // stop uploads
      this.dataSocket.dataListener.bytesTotal      = 0;                          // stop downloads
    }

    this.cleanup();

    if (!this.isConnected) {
      return;
    }

    //XXX this.writeControl("ABOR");                                             // ABOR does not seem to stop the connection in most cases
    if (this.dataSocket) {                                                       // so this is a more direct approach
      this.dataSocket.kill();
    }

    this.addEventQueue("aborted");

    if (this.observer) {
      this.observer.onAbort();
    }
  },

  checkTimeout : function(id, cmd) {
    if (this.isConnected && this.networkTimeoutID == id && this.eventQueue.length && this.eventQueue[0].cmd.indexOf(cmd) != -1) {
      this.resetConnection();
    }
  },

  checkDataTimeout : function(download, id, bytes) {
    if (this.isConnected && this.transferID == id) {
      if ((download && bytes == this.dataSocket.dataListener.bytesDownloaded)
      || (!download && bytes == this.dataSocket.progressEventSink.bytesUploaded)) {
        this.resetConnection();
        return;
      }

      var self      = this;
      var nextBytes = download ? self.dataSocket.dataListener.bytesDownloaded : self.dataSocket.progressEventSink.bytesUploaded;
      var func = function() { self.checkDataTimeout(download, id, nextBytes); };
      setTimeout(func, this.networkTimeout * 1000);
    }
  },

  resetConnection : function() {
    this.legitClose = false;                                                   // still stuck on a command so
    this.unshiftEventQueue("goodbye", "", "");                                 // try to restart the connection the hard way
    this.writeControl("QUIT");

    if (this.dataSocket) {
      this.dataSocket.kill();
    }

    this.kill();
  },

  keepAlive : function() {
    if (this.isConnected && this.keepAliveMode && this.eventQueue.length == 0) {
      this.addEventQueue("NOOP");
      this.writeControl();
    }

    var self = this;
    var func = function() { self.keepAlive(); };
    setTimeout(func, 60000);
  },

  addEventQueue : function(cmd, parameter, callback) {                           // this just creates a new queue item
    this.eventQueue.push(   { cmd: cmd, parameter: parameter || '', callback: callback || '' });
  },

  unshiftEventQueue : function(cmd, parameter, callback) {                       // ditto
    this.eventQueue.unshift({ cmd: cmd, parameter: parameter || '', callback: callback || '' });
  },

  beginCmdBatch : function() {
    this.doingCmdBatch = true;
  },

  writeControlWrapper : function() {
    if (!this.doingCmdBatch) {
      this.writeControl();
    }
  },

  endCmdBatch : function() {
    this.doingCmdBatch = false;
    this.writeControl();
  },

  writeControl : function(cmd) {
    try {
      if (!this.isReady || (!cmd && !this.eventQueue.length)) {
        return;
      }

      var parameter;
      var callback;

      if (!cmd) {
        cmd       = this.eventQueue[0].cmd;
        parameter = this.eventQueue[0].parameter;
        callback  = this.eventQueue[0].callback;
      }

      while (cmd == "aborted" || cmd == "goodbye"                                // these are sort of dummy values
         || (cmd == "TYPE" && this.transferMode      == parameter)               // or if we ignore TYPE if it's unnecessary
         || (cmd == "CWD"  && this.currentWorkingDir == parameter)) {            // or if we ignore CWD if it's unnecessary

        if ((cmd == "TYPE" && this.transferMode      == parameter)
         || (cmd == "CWD"  && this.currentWorkingDir == parameter)) {
          this.trashQueue.push(this.eventQueue[0]);
        }

        this.eventQueue.shift();

        if (this.eventQueue.length) {
          cmd       = this.eventQueue[0].cmd;
          parameter = this.eventQueue[0].parameter;
          callback  = this.eventQueue[0].callback;
        } else {
          return;
        }
      }

      this.isReady          = false;

      if (this.observer) {
        this.observer.onIsReadyChange(false);
      }

      if (!this.passiveMode && cmd == "PASV") {                                  // active mode
        cmd                   = this.ipType == "IPv4" ? "PORT" : "EPRT";
        var security          = this.security && this.protPSuccess;
        var proxy             = { proxyType: this.proxyType, proxyHost: this.proxyHost, proxyPort: this.proxyPort };
        this.dataSocket       = new ftpDataSocketMozilla(security, proxy);
        this.dataSocket.debug = this.debug;
        this.dataSocket.error = this.error;

        var activeInfo        = {};
        activeInfo.cmd        = this.eventQueue[1].cmd;
        activeInfo.ipType     = this.ipType;

        if (this.eventQueue[1].cmd        == "RETR") {
          activeInfo.localPath    = this.eventQueue[1].callback;
          activeInfo.totalBytes   = callback;
        } else if (this.eventQueue[1].cmd == "REST") {
          activeInfo.localPath    = this.eventQueue[2].callback;
          activeInfo.totalBytes   = callback;
          activeInfo.partialBytes = this.eventQueue[1].parameter;
        } else if (this.eventQueue[1].cmd == "STOR") {
          activeInfo.localPath    = this.eventQueue[1].callback;
        } else if (this.eventQueue[1].cmd == "APPE") {
          activeInfo.localPath    = this.eventQueue[1].callback.localPath;
          activeInfo.partialBytes = this.eventQueue[1].callback.remoteSize;
        }

        parameter = this.dataSocket.createServerSocket(activeInfo);
      }

      if (this.passiveMode && cmd == "PASV" && this.ipType != "IPv4") {
        cmd = "EPSV";
      }

      if (cmd == "LIST") {                                                       // don't include path in list command - breaks too many things
        parameter = this.hiddenMode ? "-al" : "";
      }

      var outputData = cmd + (parameter ? (' ' + parameter) : '') + "\r\n";      // le original bug fix! - thanks to devin
      this.controlOutstream.write(outputData, outputData.length);                // write!

      ++this.networkTimeoutID;                                                   // this checks for timeout
      var self           = this;
      var currentTimeout = this.networkTimeoutID;
      var func           = function() { self.checkTimeout(currentTimeout, cmd); };
      setTimeout(func, this.networkTimeout * 1000);

      if (cmd == "RETR" || cmd == "STOR" || cmd == "APPE") {
        ++this.transferID;
        var currentId    = this.transferID;
        var func         = function() { self.checkDataTimeout(cmd == "RETR", currentId, 0); };
        setTimeout(func, this.networkTimeout * 1000);
      }

      outputData = cmd + (parameter ? (' ' + parameter) : '');                   // write it out to the log

      if (cmd != "PASS") {
        this.appendLog("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"      + outputData,                           'output', "info");
      } else {
        this.appendLog("&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;PASS " + this.passNotShown, 'output', "info");
      }

    } catch(ex) {
      this.debug(ex);
      this.error(this.errorConnectStr);
    }
  },

  readControl : function(buffer) {
    if ((buffer == "2" && !this.isConnected) || buffer == "\r\n" || buffer == "\n") {
      return;
    }

    var lastLineOfBuffer = buffer.indexOf("\r\n") != -1 ? buffer.split("\r\n") : buffer.split("\n");
    lastLineOfBuffer     = lastLineOfBuffer.filter(this.removeBlanks);

    if (buffer != "2") {                                                         // "2"s are self-generated fake messages
      for (var x = 0; x < lastLineOfBuffer.length; ++x) {                        // add response to log
        var message   = lastLineOfBuffer[x].charAt(lastLineOfBuffer[x].length - 1) == '\r'
                      ? lastLineOfBuffer[x].substring(0, lastLineOfBuffer[x].length - 1) : lastLineOfBuffer[x];
        var errorBlah = lastLineOfBuffer[x].charAt(0) == '4' || lastLineOfBuffer[x].charAt(0) == '5';
        this.appendLog(message, errorBlah ? 'error' : 'input', errorBlah ? "error" : "info");
      }

      ++this.networkTimeoutID;
    }

    lastLineOfBuffer = lastLineOfBuffer[lastLineOfBuffer.length - 1];            // we are only interested in what the last line says
    var returnCode;

    if ((lastLineOfBuffer.length > 3 && lastLineOfBuffer.charAt(3) == '-') || lastLineOfBuffer.charAt(0) == ' ') {
      if (this.eventQueue[0].cmd == "USER" || this.eventQueue[0].cmd == "PASS") {
        this.welcomeMessage += buffer;                                           // see if the message is finished or not
      }

      return;
    } else {
      returnCode = parseInt(lastLineOfBuffer.charAt(0));                         // looks at first number of number code
    }

    var cmd;  var parameter;    var callback;

    if (this.eventQueue.length) {
      cmd       = this.eventQueue[0].cmd;
      parameter = this.eventQueue[0].parameter;
      callback  = this.eventQueue[0].callback;

      if (cmd != "LIST"  && cmd != "RETR"  && cmd != "STOR"  && cmd != "APPE"    // used if we have a loss in connection
       && cmd != "LIST2" && cmd != "RETR2" && cmd != "STOR2" && cmd != "APPE2") {
        var throwAway = this.eventQueue.shift();

        if (throwAway.cmd != "USER"    && throwAway.cmd != "PASS"    && throwAway.cmd != "PWD"
         && throwAway.cmd != "welcome" && throwAway.cmd != "goodbye" && throwAway.cmd != "aborted" && throwAway.cmd != "NOOP"
         && throwAway.cmd != "REST"    && throwAway.cmd != "SIZE"    && throwAway.cmd != "PBSZ"    && throwAway.cmd != "AUTH" && throwAway.cmd != "PROT") {
          this.trashQueue.push(throwAway);
        }
      }
    } else {
      cmd = "default";                                                           // an unexpected reply - perhaps a 421 timeout message
    }

    switch (cmd) {
      case "welcome":
        this.welcomeMessage = buffer;

        if (returnCode != 2) {
          if (this.observer) {
            this.observer.onConnectionRefused();
          }

          this.cleanup();

          break;
        }

        this.isConnected       = true;                                           // good to go

        if (this.observer) {
          this.observer.onConnected();
        }

        this.isReconnecting    = false;
        this.reconnectsLeft    = parseInt(this.reconnectAttempts);               // setup reconnection settings

        if (!this.reconnectsLeft || this.reconnectsLeft < 1) {
          this.reconnectsLeft = 1;
        }

        this.unshiftEventQueue(  "USER", this.login, "");

        if (this.security) {
          this.unshiftEventQueue("PBSZ", "0",   "");
        }

        if (this.security == "authtls") {
          this.unshiftEventQueue("AUTH", "TLS", "");
        } else if (this.security == "authssl") {
          this.unshiftEventQueue("AUTH", "SSL", "");
        }
        break;

      case "AUTH":
        if (returnCode != 2) {
          this.error(buffer, true);
          this.isConnected = false;

          this.kill();

          return;
        } else {
          var si = this.controlTransport.securityInfo;
          si.QueryInterface(Components.interfaces.nsISSLSocketControl);
          si.StartTLS();
        }
        break;

      case "PBSZ":
        if (returnCode != 2) {
          this.error(buffer, true);
          this.isConnected = false;

          this.kill();
          return;
        }
        break;

      case "PROT":
        if (buffer.substring(0, 3) == "534" && parameter == "P") {
          this.unshiftEventQueue("PROT", "C", "");
          break;
        }

        if (returnCode != 2) {
          this.error(buffer, true);
          this.protPSuccess = false;
        } else {
          this.protPSuccess = parameter == "P";
        }
        break;

      case "USER":
      case "PASS":
        if (returnCode == 2) {
          this.welcomeMessage += buffer;

          if (this.legitClose) {
            if (this.observer) {
              this.observer.onWelcomed();
            }
          }

          if (this.observer) {
            this.observer.onLoginAccepted();
          }

          if (!this.legitClose) {
            this.recoverFromDisaster();                                          // recover from previous disaster
            break;
          }

          this.legitClose   = false;

          this.unshiftEventQueue("PWD", "", "");
        } else if (cmd == "USER" && returnCode == 3) {
          this.unshiftEventQueue("PASS", this.password, "");
        } else {
          this.cleanup();                                                        // login failed, cleanup variables
          this.error(buffer, true);
          this.isConnected = false;

          this.kill();

          return;
        }
        break;

      case "PASV":
        if (returnCode != 2) {
          this.error(buffer, true);
          break;
        }

        if (this.passiveMode) {
          var dataHost;
          var dataPort;

          if (this.ipType == "IPv4") {
            buffer           = buffer.substring(buffer.indexOf("(") + 1, buffer.indexOf(")"));
            var re           = /,/g;
            buffer           = buffer.replace(re, ".");                          // parsing the port to transfer to
            var lastDotIndex = buffer.lastIndexOf(".");
            dataPort         = parseInt(buffer.substring(lastDotIndex + 1));
            dataPort        += 256 * parseInt(buffer.substring(buffer.lastIndexOf(".", lastDotIndex - 1) + 1, lastDotIndex));
            dataHost         = buffer.substring(0, buffer.lastIndexOf(".", lastDotIndex - 1));
          } else {
            buffer           = buffer.substring(buffer.indexOf("(|||") + 4, buffer.indexOf("|)"));
            dataPort         = parseInt(buffer);
            dataHost         = this.host;
          }

          var isSecure     = this.security && this.protPSuccess;
          var proxy        = { proxyType: this.proxyType, proxyHost: this.proxyHost, proxyPort: this.proxyPort };

          this.dataSocket       = new ftpDataSocketMozilla(isSecure, proxy, dataHost, dataPort);
          this.dataSocket.debug = this.debug;
          this.dataSocket.error = this.error;

          if (this.eventQueue[0].cmd        == "LIST") {                         // do what's appropriate
            this.dataSocket.connect();
          } else if (this.eventQueue[0].cmd == "RETR") {
            this.dataSocket.connect(false, this.eventQueue[0].callback,           callback);
          } else if (this.eventQueue[0].cmd == "REST") {
            this.dataSocket.connect(false, this.eventQueue[1].callback,           callback, this.eventQueue[0].parameter);
          } else if (this.eventQueue[0].cmd == "STOR") {
            this.dataSocket.connect(true,  this.eventQueue[0].callback,           0,        0);
          } else if (this.eventQueue[0].cmd == "APPE") {
            this.dataSocket.connect(true,  this.eventQueue[0].callback.localPath, 0,        this.eventQueue[0].callback.remoteSize);
          }
        }
        break;

      case "APPE":
      case "LIST":
      case "RETR":
      case "STOR":
        this.eventQueue[0].cmd = cmd + "2";

        if (returnCode == 2) {
          if (this.dataSocket.finished) {
            if (this.dataSocket.emptyFile) {                                     // XXX empty files are (still) special cases
              try {
                this.dataSocket.dataTransport.close("Finished");
              } catch (ex) {}
            }

            ++this.transferID;
            this.eventQueue.shift();
            this.trashQueue = new Array();                                       // clear the trash array, completed an 'atomic' set of operations

            if (cmd == "LIST") {
              this.listData = this.parseListData(this.dataSocket.listData, this.hiddenMode ? parameter.substring(4, parameter.length) : parameter);

              if (typeof callback == "string") {
                eval(callback);                                                  // send off list data to whoever wanted it
              } else {
                callback();
              }
            }

            break;
          } else {
            var self = this;
            var func = function() { self.readControl("2"); };
            setTimeout(func, 500);                                               // give data stream some time to finish up
            return;
          }
        }

        if (returnCode != 1) {
          this.error(buffer, true);
          this.eventQueue.shift();
          this.trashQueue = new Array();

          if (this.dataSocket) {
            this.dataSocket.kill();
          }

          break;
        }
        return;

      case "APPE2":
      case "RETR2":
      case "STOR2":
      case "LIST2":
        if (returnCode != 2) {
          if (cmd != "RETR2" && buffer.substring(0,3) != "450") {                // "450" errors are ok (i don't remember why though... :(
            this.error(buffer, true);
          }

          this.eventQueue.shift();
          this.trashQueue = new Array();

          if (this.dataSocket) {
            this.dataSocket.kill();
          }
          break;
        }

        if (this.dataSocket.finished) {
          ++this.transferID;
          this.eventQueue.shift();
          this.trashQueue = new Array();                                         // clear the trash array, completed an 'atomic' set of operations
        }

        if (cmd == "LIST2" && this.dataSocket.finished) {
          this.listData = this.parseListData(this.dataSocket.listData, this.hiddenMode ? parameter.substring(4, parameter.length) : parameter);

          if (typeof callback == "string") {
            eval(callback);                                                      // send off list data to whoever wanted it
          } else {
            callback();
          }
        } else if (!this.dataSocket.finished) {
          var self = this;
          var func = function() { self.readControl("2"); };
          setTimeout(func, 500);                                                 // give data stream some time to finish up
          return;
        }
        break;

      case "SIZE":
        if (returnCode == 2) {                                                   // used with appe commands to see where to pick up from
          var size = buffer.split(" ").filter(this.removeBlanks);
          size     = parseInt(size[1]);

          for (var x = 0; x < this.eventQueue.length; ++x)
            if (callback == this.eventQueue[x].cmd) {
              if (callback == "STOR") {
                this.eventQueue[x].cmd      = "APPE";
                this.eventQueue[x].callback = { localPath: this.eventQueue[x].callback,           remoteSize: size };
              }

              if (callback == "APPE") {
                this.eventQueue[x].callback = { localPath: this.eventQueue[x].callback.localPath, remoteSize: size };
              }

              break;
            }
        } else {                                                                 // our size command didn't work out, make sure we're not doing an APPE
          for (var x = 0; x < this.eventQueue.length; ++x) {
            if (this.eventQueue[x].cmd == "APPE") {
              this.eventQueue[x].cmd      = "STOR";
              this.eventQueue[x].callback = this.eventQueue[x].callback.localPath;
              break;
            }
          }
        }
        break;

      case "RNFR":
      case "REST":
        if (returnCode != 3) {
          if (cmd == "RNFR") {
            this.eventQueue = new Array();
            this.trashQueue = new Array();
          }

          this.error(buffer, true);                                              // should still be able to go on without this, just not with resuming
          break;
        }
        break;

      case "MKD":
      case "SITE CHMOD":
      case "RNTO":
      case "DELE":
      case "RMD":
        if (returnCode != 2) {
          this.error(buffer, true);
        } else {
          if (cmd == "RMD") {                                                    // clear out of cache if it's a remove directory
            this.removeCacheEntry(this.constructPath(this.currentWorkingDir, parameter));
          }

          if (typeof callback == "string") {
            eval(callback);                                                      // send off list data to whoever wanted it
          } else {
            callback();
          }
        }

        this.trashQueue = new Array();
        break;

      case "CWD":
        if (returnCode != 2) {                                                   // if it's not a directory
          if (callback && typeof callback == "function") {
            callback(false);
          } else if (this.observer) {
            this.observer.onDirNotFound(buffer);
            this.error(buffer, true);
          }
        } else {
          this.currentWorkingDir = parameter;

          if (this.observer) {                                                   // else navigate to the directory
            this.observer.onChangeDir(parameter, typeof callback == "boolean" ? callback : "");
          }

          if (callback && typeof callback == "function") {
            callback(true);
          }
        }
        break;

      case "PWD":                                                                // gotta check for chrooted directories
        buffer = buffer.substring(buffer.indexOf("\"") + 1, buffer.lastIndexOf("\""));              // if buffer is not '/' we're chrooted
        this.currentWorkingDir = buffer;

        if (this.observer) {
          this.observer.onChangeDir(buffer != '/' && this.initialPath == '/' ? buffer : '');
        }

        this.trashQueue = new Array();
        break;

      case "aborted":
        break;

      case "TYPE":
        if (returnCode != 2) {
          this.error(buffer, true);
        } else {
          this.transferMode = parameter;
        }
        break;
      case "goodbye":                                                            // you say yes, i say no, you stay stop...
      case "NOOP":
      default:
        if (returnCode != 2) {
          this.error(buffer, true);
        }
        break;
    }

    this.isReady = true;

    if (this.observer) {
      this.observer.onIsReadyChange(true);
    }

    if (this.eventQueue.length && this.eventQueue[0].cmd != "welcome") {         // start the next command
      this.writeControl();
    } else {
      if (this.localRefreshLater) {
        var dir                 = new String(this.localRefreshLater);
        this.localRefreshLater  = "";

        if (this.observer) {
          this.observer.onShouldRefresh(true, false, dir);
        }
      }

      if (this.remoteRefreshLater) {
        var dir                 = new String(this.remoteRefreshLater);
        this.remoteRefreshLater = "";

        if (this.observer) {
          this.observer.onShouldRefresh(false, true, dir);
        }
      }
    }
  },

  changeWorkingDirectory : function(path, callback) {
    this.addEventQueue("CWD", path, callback);
    this.writeControlWrapper();
  },

  makeDirectory : function(path, callback) {
    this.addEventQueue("CWD", path.substring(0, path.lastIndexOf('/') ? path.lastIndexOf('/') : 1), true);
    this.addEventQueue("MKD", path.substring(path.lastIndexOf('/') + 1), callback);
    this.writeControlWrapper();
  },

  remove : function(isDirectory, path, callback) {
    if (isDirectory) {
      this.unshiftEventQueue("RMD",    path.substring(path.lastIndexOf('/') + 1), callback);
      this.unshiftEventQueue("CWD",    path.substring(0, path.lastIndexOf('/') ? path.lastIndexOf('/') : 1), true);

      var self         = this;
      var listCallback = function() { self.removeRecursive(path); };
      this.list(path, listCallback, true, true);
    } else {
      this.unshiftEventQueue("DELE",   path.substring(path.lastIndexOf('/') + 1), callback);
      this.unshiftEventQueue("CWD",    path.substring(0, path.lastIndexOf('/') ? path.lastIndexOf('/') : 1), true);
    }

    this.writeControlWrapper();
  },

  removeRecursive : function(parent) {                                           // delete subdirectories and files
    var files = this.listData;

    for (var x = 0; x < files.length; ++x) {
      var remotePath = this.constructPath(parent, files[x].leafName);

      if (files[x].isDirectory()) {                                              // delete a subdirectory recursively
        this.unshiftEventQueue("RMD",  remotePath.substring(remotePath.lastIndexOf('/') + 1), "");
        this.unshiftEventQueue("CWD",  parent, true);
        this.removeRecursiveHelper(remotePath);
      } else {                                                                   // delete a file
        this.unshiftEventQueue("DELE", remotePath.substring(remotePath.lastIndexOf('/') + 1), "");
        this.unshiftEventQueue("CWD",  parent, true);
      }
    }
  },

  removeRecursiveHelper : function(remotePath) {
    var self           = this;
    var listCallback   = function() { self.removeRecursive(remotePath); };
    this.list(remotePath, listCallback, true, true);
  },

  rename : function(oldName, newName, callback, isDir) {
    if (isDir) {
      gFtp.removeCacheEntry(oldName);
    }

    this.addEventQueue("RNFR", oldName);                                         // rename the file
    this.addEventQueue("RNTO", newName, callback);
    this.writeControlWrapper();
  },

  changePermissions : function(permissions, path, callback) {
    this.addEventQueue("CWD", path.substring(0, path.lastIndexOf('/') ? path.lastIndexOf('/') : 1), true);
    this.addEventQueue("SITE CHMOD", permissions + ' ' + path.substring(path.lastIndexOf('/') + 1), callback);
    this.writeControlWrapper();
  },

  custom : function(cmd) {
    this.addEventQueue(cmd);
    this.writeControlWrapper();
  },

  list : function(path, callback, skipCache, recursive) {
    if (!skipCache && this.sessionsMode) {
      try {                                                                      // check the cache first
        var cacheSession   = this.cacheService.createSession("fireftp", 0, true);
        var cacheDesc      = cacheSession.openCacheEntry("ftp://" + this.version + this.connectedHost + path,
                                                         Components.interfaces.nsICache.ACCESS_READ, false);

        if (cacheDesc.dataSize) {
          var cacheIn       = cacheDesc.openInputStream(0);
          var cacheInstream = Components.classes["@mozilla.org/binaryinputstream;1"].createInstance(Components.interfaces.nsIBinaryInputStream);
          cacheInstream.setInputStream(cacheIn);
          this.listData     = cacheInstream.readBytes(cacheInstream.available());
          this.listData     = eval(this.listData);
          cacheInstream.close();
          cacheDesc.close();

          if (typeof callback == "string") {
            eval(callback);                                                  // send off list data to whoever wanted it
          } else {
            callback();
          }

          return;
        }

        cacheDesc.close();
      } catch (ex) { }
    }

    if (recursive) {
      this.unshiftEventQueue(  "LIST", (this.hiddenMode ? "-al " : '') + path, callback);
      this.unshiftEventQueue(  "PASV",  "", "");
      this.unshiftEventQueue(  "CWD", path, "");

      if (this.security) {
        this.unshiftEventQueue("PROT", "P", "");
      }

      this.unshiftEventQueue(  "TYPE", "A", "");
    } else {
      this.addEventQueue(      "TYPE", "A", "");

      if (this.security) {
        this.addEventQueue(    "PROT", "P", "");
      }

      this.addEventQueue(      "CWD", path, "");
      this.addEventQueue(      "PASV",  "", "");
      this.addEventQueue(      "LIST", (this.hiddenMode ? "-al " : '') + path, callback);
    }

    this.writeControlWrapper();
  },

  parseListData : function(data, path) {
    /* Unix style:                     drwxr-xr-x 1 user01 ftp 512 Jan 29 23:32 prog
     * Alternate Unix style:           drwxr-xr-x 1 user01 ftp 512 Jan 29 1997  prog
     * Alternate Unix style:           drwxr-xr-x 1 1      1   512 Jan 29 23:32 prog
     * A symbolic link in Unix style:  lrwxr-xr-x 1 user01 ftp 512 Jan 29 23:32 prog -> prog2000
     * MS-DOS style:                   01-29-97 11:32PM <DIR> prog
     * Novell style:                   drwxr-xr-x 1 user01     512 Jan 29 23:32 prog
     * Weird style:                    drwxr-xr-x 1 user5424867    Jan 29 23:32 prog, where 5424867 is the size
     * OS/2 style:                     0           DIR 01-29-97  23:32  PROG
     * OS/2 style:                     2243        RA  04-05-103 00:22  PJL
     * OS/2 style:                     60              11-18-104 06:54  chkdsk.log
     * AIX style:                      -rw-r--r-- 1 sys   sys   30 05 Nov 2003  foo.txt
     */

    this.debug(data.replace(/</g, '&lt;').replace(/>/g, '&gt;'), "INFO");

    var items   = data.indexOf("\r\n") != -1 ? data.split("\r\n") : data.split("\n");
    items       = items.filter(this.removeBlanks);
    var curDate = new Date();

    if (items.length) {                                                          // some ftp servers send 'count <number>' or 'total <number>' first
      if (items[0].indexOf("count") == 0 || items[0].indexOf("total") == 0 || items[0].split(" ").filter(this.removeBlanks).length == 2) {
        items.shift();                                                           // could be in german or croatian or what have you
      }
    }

    for (var x = 0; x < items.length; ++x) {
      if (!items[x]) {                                                           // some servers put in blank lines b/w entries, aw, for cryin' out loud
        items.splice(x, 1);
        --x;
        continue;
      }

      var temp = items[x];                                                       // account for collisions:  drwxr-xr-x1017 user01

      if (!parseInt(items[x].charAt(0)) && items[x].charAt(0) != '0' && items[x].charAt(10) != ' ') {     // this is mimicked below if weird style
        items[x] = items[x].substring(0, 10) + ' ' + items[x].substring(10, items[x].length);
      }

      items[x]   = items[x].split(" ").filter(this.removeBlanks);

      if (!parseInt(items[x][0].charAt(0)) && items[x][0].charAt(0) != '0')  {   // unix style - so much simpler with you guys
        var offset = 0;

        if (items[x][3].search(this.remoteMonths) != -1 && items[x][5].search(this.remoteMonths) == -1) { // added to support weird servers
          var weird = temp;

          if (weird.charAt(10) != ' ') {                                         // same as above code
            weird = weird.substring(0, 10) + ' ' + weird.substring(10, weird.length);
          }

          var weirdIndex = 0;

          for (var y = 0; y < items[x][2].length; ++y) {
            if (parseInt(items[x][2].charAt(y))) {
              weirdIndex = weird.indexOf(items[x][2]) + y;
              break;
            }
          }

          weird    = weird.substring(0, weirdIndex) + ' ' + weird.substring(weirdIndex, weird.length);

          items[x] = weird.split(" ").filter(this.removeBlanks);
        }

        if (items[x][4].search(this.remoteMonths) != -1) {                       // added to support novell servers
          offset   = 1;
        }

        var index = 0;
        for (var y = 0; y < 7 - offset; ++y) {
          index = temp.indexOf(items[x][y], index) + items[x][y].length + 1;
        }

        var name    = temp.substring(temp.indexOf(items[x][7 - offset], index) + items[x][7 - offset].length + 1, temp.length);
        name        = name.substring(name.search(/[^\s]/));
        var symlink = "";

        if (items[x][0].charAt(0) == 'l') {
          symlink   = name;
          name      = name.substring(0, name.indexOf("->") - 1);
          symlink   = symlink.substring(symlink.indexOf("->") + 3);
        }

        name           = (name.lastIndexOf('/') == -1 ? name : name.substring(name.lastIndexOf('/') + 1));
        var remotepath = this.constructPath(path, name);
        var month      = this.l10nMonths[this.remoteMonths.search(items[x][5 - offset]) / 4];
        items[x]       = { permissions : items[x][0],
                           hardLink    : items[x][1],
                           user        : items[x][2],
                           group       : (offset ? "" : items[x][3]),
                           fileSize    : items[x][4 - offset],
                           date        : month + ' ' + items[x][6 - offset] + ' ' + items[x][7 - offset],
                           leafName    : name,
                           isDir       : items[x][0].charAt(0) == 'd',
                           isDirectory : function() { return this.isDir },
                           isSymlink   : items[x][0].charAt(0) == 'l',
                           symlink     : symlink,
                           path        : remotepath };

      } else if (items[x][0].indexOf('-') == -1) {                               // os/2 style
        var offset = 0;

        if (items[x][2].indexOf(':') != -1) {                                    // if "DIR" and "A" are missing
          offset   = 1;
        }

        var rawDate    = items[x][2 - offset].split("-");
        var rawTime    = items[x][3 - offset];
        var timeOrYear = rawTime;
        rawTime        = rawTime.split(":");

        for (var y = 0; y < rawDate.length; ++y) {
          rawDate[y]   = parseInt(rawDate[y], 10);                               // leading zeros are treated as octal so pass 10 as base argument
        }

        for (var y = 0; y < rawTime.length; ++y) {
          rawTime[y]   = parseInt(rawTime[y], 10);
        }

        rawDate[2]     = rawDate[2] + 1900;                                      // ah, that's better
        var parsedDate = new Date(rawDate[2], rawDate[0] - 1, rawDate[1], rawTime[0], rawTime[1]);  // month-day-year format

        if (new Date() - parsedDate > 15600000000) {                             // roughly 6 months
          timeOrYear   = rawDate[2];
        }

        var month      = this.l10nMonths[parsedDate.getMonth()];
        var name       = temp.substring(temp.indexOf(items[x][3 - offset]) + items[x][3 - offset].length + 1, temp.length);
        name           = name.substring(name.search(/[^\s]/));
        name           = (name.lastIndexOf('/') == -1 ? name : name.substring(name.lastIndexOf('/') + 1));
        items[x]       = { permissions : items[x][1] == "DIR" ? "d---------" : "----------",
                           hardLink    : "",
                           user        : "",
                           group       : "",
                           fileSize    : items[x][0],
                           date        : month + ' ' + rawDate[1] + ' ' + timeOrYear,
                           leafName    : name,
                           isDir       : items[x][1] == "DIR",
                           isDirectory : function() { return this.isDir },
                           isSymlink   : false,
                           symlink     : "",
                           path        : this.constructPath(path, name) };

      } else {                                                                   // ms-dos style
        var rawDate    = items[x][0].split("-");
        var rawTime    = items[x][1].substring(0, 5);                            // get rid of PM, AM
        var timeOrYear = rawTime;
        rawTime        = rawTime.split(":");

        for (var y = 0; y < rawDate.length; ++y) {
          rawDate[y]   = parseInt(rawDate[y], 10);
        }

        for (var y = 0; y < rawTime.length; ++y) {
          rawTime[y]   = parseInt(rawTime[y], 10);
        }

        if (rawDate[2] < 70) {                                                   // assuming you didn't have some files left over from 1904
          rawDate[2]   = rawDate[2] + 2000;                                      // ah, that's better
        }

        var parsedDate = new Date(rawDate[2], rawDate[0] - 1, rawDate[1], rawTime[0], rawTime[1]);  // month-day-year format

        if (new Date() - parsedDate > 15600000000) {                             // roughly 6 months
          timeOrYear   = rawDate[2];
        }

        var month      = this.l10nMonths[parsedDate.getMonth()];
        var name       = temp.substring(temp.indexOf(items[x][2], temp.indexOf(items[x][1]) + items[x][1].length + 1)
                         + items[x][2].length + 1, temp.length);
        name           = name.substring(name.search(/[^\s]/));
        name           = (name.lastIndexOf('/') == -1 ? name : name.substring(name.lastIndexOf('/') + 1));
        items[x]       = { permissions : items[x][2] == "<DIR>" ? "d---------" : "----------",
                           hardLink    : "",
                           user        : "",
                           group       : "",
                           fileSize    : items[x][2] == "<DIR>" ? '0' : items[x][2],
                           date        : month + ' ' + rawDate[1] + ' ' + timeOrYear,
                           leafName    : name,
                           isDir       : items[x][2] == "<DIR>",
                           isDirectory : function() { return this.isDir },
                           isSymlink   : false,
                           symlink     : "",
                           path        : this.constructPath(path, name) };
      }

      var dateTemp  = items[x].date;                                             // this helps with sorting by date
      var dateMonth = dateTemp.substring(0, 3);
      var dateIndex = this.l10nMonths.indexOf(dateMonth);
      dateTemp      = this.remoteMonths.substr(dateIndex * 4, 3) + dateTemp.substring(3);

      if (items[x].date.indexOf(':') != -1) {
        dateTemp = dateTemp + ' ' + (curDate.getFullYear() - (curDate.getMonth() < dateIndex ? 1 : 0));
      }

      items[x].lastModifiedTime = Date.parse(dateTemp);
    }

    var directories = new Array();                                               // sort directories to the top
    var files       = new Array();

    for (var x = 0; x < items.length; ++x) {
      if (items[x].isDirectory()) {
        directories.push(items[x]);
      } else {
        files.push(items[x]);
      }
    }

    for (var x = directories.length - 1; x >= 0; --x) {                          // get rid of "." or ".."
      if (directories[x].leafName == "." || directories[x].leafName == "..") {   // this can screw up things on recursive deletions
        directories.splice(x, 1);
      }
    }

    items = directories.concat(files);

    if (this.sessionsMode) {
      try {                                                                      // put in cache
        var cacheSession = this.cacheService.createSession("fireftp", 0, true);
        var cacheDesc    = cacheSession.openCacheEntry("ftp://" + this.version + this.connectedHost + path,
                                                       Components.interfaces.nsICache.ACCESS_WRITE, false);
        var cacheOut     = cacheDesc.openOutputStream(0);
        var cacheData    = items.toSource();
        cacheOut.write(cacheData, cacheData.length);
        cacheOut.close();
        cacheDesc.close();
      } catch (ex) {
        this.debug(ex);
      }
    }

    return items;
  },

  download : function(remotePath, localPath, remoteSize, resume, localSize) {
    this.addEventQueue(  "CWD",  remotePath.substring(0, remotePath.lastIndexOf('/') ? remotePath.lastIndexOf('/') : 1), true);

    this.addEventQueue(  "TYPE", this.detectAscii(remotePath));

    if (this.security) {
      this.addEventQueue("PROT", "P");
    }

    this.addEventQueue(  "PASV", "", remoteSize);

    if (resume) {
      this.addEventQueue("REST", localSize);
    }

    this.addEventQueue(  "RETR", remotePath.substring(remotePath.lastIndexOf('/') + 1), localPath);

    this.writeControlWrapper();
  },

  upload : function(localPath, remotePath, resume, remoteSize) {
    this.addEventQueue(  "CWD",  remotePath.substring(0, remotePath.lastIndexOf('/') ? remotePath.lastIndexOf('/') : 1), true);

    var leafName = remotePath.substring(remotePath.lastIndexOf('/') + 1);

    if (resume) {
      this.addEventQueue("SIZE", leafName, "APPE");  // need to do a size check
    }

    this.addEventQueue(  "TYPE", this.detectAscii(remotePath));

    if (this.security) {
      this.addEventQueue("PROT", "P");
    }

    this.addEventQueue(  "PASV");

    if (resume) {
      this.addEventQueue("APPE", leafName, { localPath: localPath, remoteSize: remoteSize } );
    } else {
      this.addEventQueue("STOR", leafName,   localPath);
    }

    this.writeControlWrapper();
  },

  removeCacheEntry : function(path) {
    try {
      var cacheSession = this.cacheService.createSession("fireftp", 0, true);
      var cacheDesc    = cacheSession.openCacheEntry("ftp://" + this.version + this.connectedHost + path,
                                                     Components.interfaces.nsICache.ACCESS_WRITE, false);
      cacheDesc.doom();
      cacheDesc.close();
    } catch (ex) {
      this.debug(ex);
    }
  },

  isListing : function() {                                                       // check queue to see if we're listing
    for (var x = 0; x < this.eventQueue.length; ++x) {
      if (this.eventQueue[x].cmd.indexOf("LIST") != -1) {
        return true;
      }
    }

    return false;
  },

  recoverFromDisaster : function() {                                             // after connection lost, try to restart queue
    if (this.eventQueue.length && this.eventQueue[0].cmd == "goodbye") {
      this.eventQueue.shift();
    }

    if (this.eventQueue.cmd) {
      this.eventQueue = new Array(this.eventQueue);
    }

    if (this.eventQueue.length && (this.eventQueue[0].cmd == "LIST" || this.eventQueue[0].cmd == "LIST2"
                               ||  this.eventQueue[0].cmd == "RETR" || this.eventQueue[0].cmd == "RETR2"
                               ||  this.eventQueue[0].cmd == "REST" || this.eventQueue[0].cmd == "APPE"
                               ||  this.eventQueue[0].cmd == "STOR" || this.eventQueue[0].cmd == "STOR2"
                               ||  this.eventQueue[0].cmd == "PASV" || this.eventQueue[0].cmd == "APPE2")) {
      var cmd       = this.eventQueue[0].cmd;
      var parameter = this.eventQueue[0].parameter;
      if (cmd == "LIST2" || cmd == "RETR2" || cmd == "STOR2" || cmd == "APPE2") {
        this.eventQueue[0].cmd = this.eventQueue[0].cmd.substring(0, 4);
      }

      cmd = this.eventQueue[0].cmd;

      if (cmd == "REST") {                                                       // set up resuming for these poor interrupted transfers
        try {
          var file = Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile);
          file.initWithPath(this.eventQueue[1].callback);

          if (file.fileSize) {
            this.eventQueue[0].parameter = file.fileSize;
          }
        } catch (ex) {
          this.debug(ex);
        }
      } else if (cmd == "RETR") {
        try {
          var file = Components.classes['@mozilla.org/file/local;1'].createInstance(Components.interfaces.nsILocalFile);
          file.initWithPath(this.eventQueue[0].callback);

          if (file.fileSize) {
            this.unshiftEventQueue("REST", file.fileSize, "");
          }
        } catch (ex) {
          this.debug(ex);
        }
      }

      for (var x = this.trashQueue.length - 1; x >= 0; --x) {                    // take cmds out of the trash and put them back in the eventQueue
        this.eventQueue.unshift(this.trashQueue[x]);
      }

      if (cmd == "STOR" || cmd == "APPE") {                                      // more resuming fun - this time for the stor/appe commandds
        this.unshiftEventQueue("SIZE", parameter, cmd);
      }
    } else if (this.eventQueue.length && this.eventQueue[0].cmd == "RNTO" && this.trashQueue[this.trashQueue.length - 1].cmd == "RNFR") {
      this.unshiftEventQueue(this.trashQueue[this.trashQueue.length - 1]);
    }

    if (this.currentWorkingDir) {
      this.unshiftEventQueue("CWD", this.currentWorkingDir, true);
      this.currentWorkingDir = "";
    }

    this.trashQueue = new Array();
  },

  detectAscii : function(path) {                                                 // detect an ascii file - returns "A" or "I"
    if (this.fileMode == 1) {                                                    // binary
      return "I";
    }

    if (this.fileMode == 2) {                                                    // ASCII
      return "A";
    }

    path = path.substring(path.lastIndexOf('.') + 1);                            // manually detect

    for (var x = 0; x < this.asciiFiles.length; ++x) {
      if (this.asciiFiles[x].toLowerCase() == path.toLowerCase()) {
        return "A";
      }
    }

    return "I";
  },

  constructPath : function(parent, leafName) {
    return parent + (parent.charAt(parent.length - 1) != '/' ? '/' : '') + leafName;
  },

  removeBlanks : function(element, index, array) {
    return element;
  }
};
